#define KNOCK_GAIN_FLAGS_GAIN_CHANGED	0x01
#define KNOCK_GAIN_FLAGS_LOW_GAIN_MODE	0x10

#define KNOCK_CALC_FLAGS_CONTROLLABLE_CONDITIONS	0x40
#define KNOCK_CALC_FLAGS_FULL_SENSITIVITY		0x80
#define KNOCK_CALC_FLAGS_LOW_KNOCK_SUPRESSION		0x0200
#define KNOCK_CALC_FLAGS_SENSOR_TESTABLE		0x0400

#define GPIO3_LOW_GAIN	0x02

void knock_window_close_call()
{
	if (!(mode_flags1 & MODE_FLAGS1_KNOCK_CONTROL_ENABLED)) return;

	adc_convert_knock();

	uint_fast16_t l_adck = adc_byte_knock;
	uint_fast16_t l_sig;

#ifndef WIDEBAND_KNOCK_SENSOR
	if (100 < l_adck) {
		l_sig = 55 + ps_scale_base256(l_adck - 100, 435);
	} else if (92 < l_adck) {
		l_sig = 49 + ps_scale_base256(l_adck - 92, 192)
	} else {
		l_sig = ps_scale_base256(l_adck, 136);
	}
#else
#error "TODO: write new knock sensor translation to knock_signal"
#endif
	uint_fast16_t l_old_signal = knock_signal;
	knock_signal = MIN16(l_sig, 255);

	if ((knock_calc_flags & KNOCK_CALC_FLAGS_SENSOR_TESTABLE)
	&& !(knock_gain_flags & KNOCK_GAIN_FLAGS_GAIN_CHANGED)) {
		/*knock sensor diagnostics*/
	} else {
		knock_dynamics = 0;
		knock_free_dynamics = 0;
		mut_knock_change = 0;
	}

	uint_fast16_t knock_positive = s_sub16(knock_signal, knock_base);
	uint_fast16_t l_knock_add = 0;

	if (!(knock_calc_flags & KNOCK_CALC_FLAGS_CONTROLLABLE_CONDITIONS)) {
		knock_sum = 0;
		reset_low_knock_state();
	} else if (combustion_faults & 0x01) { //knock sensor fault
		knock_sum = knock_calc_flags & KNOCK_CALC_FLAGS_FULL_SENSITIVITY
			? flash_knock_sensor_fault_knock_sum_full_sensitivity
			: flash_knock_sensor_fault_knock_sum_reduced_sensitivity;
		reset_low_knock_state();
	} else {
		l_knock_add = knock_positive
			? MIN16(255, ps_mul_divu16(knock_positive,
				knock_calc_flags & KNOCK_CALC_FLAGS_FULL_SENSITIVITY
					? flash_knock_positive_to_addition_full_sensitivity //32
					: flash_knock_positive_to_addition_reduced_sensitivity, //16
				knock_base * 8) + 1)
			: 0;
#define LOW_KNOCK_INVALID_CYLINDER 0x8000
		if (knock_calc_flags & KNOCK_CALC_FLAGS_LOW_KNOCK_SUPRESSION) {
			if (++low_knock_cylinder1_counter >= low_knock_counter_threshold) {
				low_knock_cylinder1_counter = low_knock_counter_threshold;
				low_knock_cylinder1 = LOW_KNOCK_INVALID_CYLINDER;
			}
			if (++low_knock_cylinder2_counter >= low_knock_counter_threshold) {
				low_knock_cylinder2_counter = low_knock_counter_threshold;
				low_knock_cylinder2 = LOW_KNOCK_INVALID_CYLINDER;
			}
			rolling_low_knock_flag <<= 1;
			if ((l_knock_add >= flash_low_knock_min_knock) //1
			&& (low_knock_threshold >= l_knock_add)) { //2
				rolling_low_knock_flag |= 0x01;
				if (spark_active_cylinder == low_knock_cylinder1) {
					low_knock_cylinder1_counter = 0;
					l_knock_add = 0;
				} else if (spark_active_cylinder == low_knock_cylinder2) {
					low_knock_cylinder2_counter = 0;
					l_knock_add = 0;
				} else if (!flash_low_knock_register_cylinder_on_second_run
					|| (rolling_low_knock_flag & 0x10)) {//this means that last time it was low knock too
				} else {
					if (low_knock_cylinder1 == LOW_KNOCK_INVALID_CYLINDER) {
						low_knock_cylinder1 = spark_active_cylinder;
						low_knock_cylinder1_counter = 0;
						if (!flash_low_knock_register_cylinder_on_second_run) l_knock_add = 0;
					} else if (low_knock_cylinder2 == LOW_KNOCK_INVALID_CYLINDER) {
						low_knock_cylinder2 = spark_active_cylinder;
						low_knock_cylinder2_counter = 0;
						if (!flash_low_knock_register_cylinder_on_second_run) l_knock_add = 0;
					}
				}
			}
		} else {
			reset_low_knock_state();
		}

		knock_sum_max = knock_sum_max_source;
		l_knock_add = MIN16(l_knock_add, knock_calc_flags & KNOCK_CALC_FLAGS_FULL_SENSITIVITY ? flash_knock_full_sensitivitity_knock_addition_max : flash_knock_reduced_sensitivitity_knock_addition_max);
		knock_sum = MIN16(knock_sum + l_knock_add, knock_sum_max);
	}

	mut_knock_addition = l_knock_add;
	mut_knock_signal_time_averaged = knock_signal_time_averaged;
	mut_knock_base = knock_base;

	update_knock_signal_spark_averaged();

	update_knock_gain_flags_and_variables();

}

//compiled as part of KWCC
static inline void update_knock_gain_flags_and_variables()
{
	/*wideband knock sensor has x1,x2,x4 gains, resonant type x1,x3*/
	if (!(gpio3 & GPIO3_LOW_GAIN)) {
		if (flash_knock_gain_switch_up_knock_base_min <= knock_base) {
			knock_gain_flags |= KNOCK_GAIN_FLAGS_LOW_GAIN_MODE;
			knock_gain_switched_at_rpm = engine_rpm_8kmax;
		}
	} else if ((engine_rpm_8kmax < s_sub16(knock_gain_switched_at_rpm, flash_knock_gain_switch_rpm_hysteresis))
	|| (knock_base < flash_knock_gain_switch_down_knock_base_max)) {
		}
		knock_gain_flags &= ~KNOCK_GAIN_FLAGS_LOW_GAIN_MODE; 
	}

	knock_gain_flags &= ~KNOCK_GAIN_FLAGS_GAIN_CHANGED;

	if (knock_gain_flags & KNOCK_GAIN_FLAGS_LOW_GAIN_MODE) {
		if (!(gpio3 & GPIO3_LOW_GAIN)) {
			knock_signal_spark_averaged = ps_divu16(knock_signal_spark_averaged, 3);
			knock_signal_time_averaged = ps_divu16(knock_signal_time_averaged, 3);
			knock_base = mid16(ps_divu16(knock_base, 3), 1, 255);
			knock_gain_flags |= KNOCK_GAIN_FLAGS_GAIN_CHANGED;
		}
		gpio3 |= GPIO3_LOW_GAIN;
	} else {
		if (gpio3 & GPIO3_LOW_GAIN) {
			knock_signal_spark_averaged = s_mul16(knock_signal_spark_averaged, 3);
			knock_signal_time_averaged = s_mul16(knock_signal_time_averaged, 3);
			knock_base = s_mul16(knock_base, 3);
			knock_gain_flags |= KNOCK_GAIN_FLAGS_GAIN_CHANGED;
		}
		gpio3 &= GPIO3_LOW_GAIN;
	}

	knock_update_gain();
#ifdef WIDEBAND_KNOCK_SENSOR
	if (flash_knock_enable_bandpass_update)
		knock_update_bandpass_frequency();
#endif

}

//compiled as part of KWCC
static inline void update_knock_signal_spark_averaged()
{
	uint_fast16_t l_mode = 0x20;

	if (knock_signal >= (knock_signal_time_averaged / 256)) {
		S_INC_U16(knock_above_base_counter);
	} else {
		knock_above_base_counter = 0;
	}

	if (decays_x1.knock_rpm_accel) {
		l_mode = 0x0100;
	} else if (knock_above_base_counter >= flash_knock_above_base_threshold) {
		l_mode = 0x10;
	} else {
		uint_fast16_t l_ksta = knock_signal_time_averaged / 256;
		if (l_ksta && (knock_signal * 8 >= flash_knock_above_base_spike * l_ksta)) {
			l_mod = 0x0100;
		}
	}
	uint_fast16_t l_rate = 0;
	uint_fast16_t l_add;
	if (l_mode == 0x10) {
		if (flash_knock_integral_smoothing_rate1) { //16
			l_rate = (flash_knock_integral_smoothing_rate1 << 16 - 0x10000) / flash_knock_integral_smoothing_rate1;
			l_add = s16_divu32_16(mul16(knock_signal, 256), flash_knock_integral_smoothing_rate1);
		}
	} else if (l_mode == 0x20) {
		if (flash_knock_integral_smoothing_rate2) { //32
			l_rate = (flash_knock_integral_smoothing_rate2 << 16 - 0x10000) / flash_knock_integral_smoothing_rate2;
			l_add = s16_divu32_16(mul16(knock_signal, 256), flash_knock_integral_smoothing_rate2);
		}
	} else {
		if (flash_knock_integral_smoothing_rate3) { //255
			l_rate = (flash_knock_integral_smoothing_rate3 << 16 - 0x10000) / flash_knock_integral_smoothing_rate3;
			l_add = s16_divu32_16(mul16(knock_signal, 256), flash_knock_integral_smoothing_rate3);
		}
	}
	knock_signal_spark_averaged = p_conv32_16(s_add32(mul16(knock_signal_spark_averaged, l_rate), conv16to32(l_add)));

}
